{-# LANGUAGE GADTs                 #-}
{-# LANGUAGE DataKinds             #-}
{-# LANGUAGE TypeFamilies          #-}
{-# LANGUAGE TypeOperators         #-}
{-# LANGUAGE StandaloneDeriving    #-}
{-# LANGUAGE FlexibleInstances     #-}
{-# LANGUAGE ScopedTypeVariables   #-}
{-# LANGUAGE NoImplicitPrelude     #-}
{-# LANGUAGE UndecidableInstances  #-}
{-# LANGUAGE TemplateHaskell       #-}
{-# LANGUAGE QuasiQuotes           #-}
{-# LANGUAGE PolyKinds             #-}
{-# LANGUAGE UnicodeSyntax         #-}

-- | Defines a vector type, analogous to a list, that achieves
-- type-safe operation like resizing and concatenation by using
-- new dependantly typed language features in haskell
module Data.Vec where

import Relation.Equality (gcastWith, cong, sym)
import Data.Nat
import Data.Bool
import Data.Char         (Char)
import Data.Function     ((∘), ($))
import Data.TypeClass    (Show, Eq, Ord, Num, IsList, IsString)
import Data.Singletons   (SingI, sing)

import qualified Data.TypeClass as T


infixr 5 :-

-- | The 'Vec' datatype
data Vec (a ∷ ★) ∷ ℕ → ★ where
  Nil  ∷ Vec a Z                   -- ^ empty 'Vec', length 0
  (:-) ∷ a → Vec a n → Vec a (S n) -- ^ "cons" operator


-- | 'String' type alias for vector of characters
type String n = Vec Char n


deriving instance Eq a ⇒ Eq (Vec a n)

instance Show a ⇒ Show (Vec a n) where
  showsPrec d = T.showsPrec d ∘ toList

instance SingI n ⇒ IsList (Vec a n) where
  type Item (Vec a n) = a
  fromList = fromList
  toList   = toList

instance SingI n ⇒ IsString (String n) where
  fromString = fromList



-- * Conversions

-- | Convert a 'Vec' into a List
toList ∷ Vec a n → [a]
toList Nil = []
toList (x :- xs) = x : toList xs


-- | Convert a List into a 'Vec'
--
-- It's not possible to convert an infinite list to a vector
-- as haskell does not permit constructing infinite type,
-- which the resulting vector length would be.
fromList ∷ SingI n ⇒ [a] → Vec a n
fromList = f sing
  where  f ∷ Sℕ n → [a] → Vec a n
         f SZ     _      = Nil
         f (SS n) (x:xs) = x :- f n xs



-- * Basic functions

-- | Vector concatenation
(⧺) ∷ Vec a n → Vec a m → Vec a (n :+ m)
(⧺) (x :- xs) ys = x :- xs ⧺ ys
(⧺) Nil       ys = ys


-- | Extracts the 'head' (ie the first element) of a nonempty 'Vec'.
head ∷ Vec a (S n) → a
head (x :- _) = x


-- | Extracts the last element of a nonempty 'Vec'.
last ∷ Vec a (S n) → a
last (x :- Nil)     = x
last (x :- y :- ys) = last (y :- ys)


-- | Applied to a nonempty 'Vec' returns everything but its first element.
-- > x ≡ head x ⧺ tail x
tail ∷ Vec a (S n) → Vec a n
tail (_ :- xs) = xs


-- | Applied to a nonempty 'Vec' returns everything but its last element.
-- > x ≡ init x ⧺ last x
init ∷ Vec a (S n) → Vec a n
init (x :- Nil)     = Nil
init (x :- y :- ys) = x :- init (y :- ys)


-- | Given a nonempty 'Vec' returns the first element
-- and the rest of the vector in a tuple.
-- > uncons x ≡ (head x, tail x)
uncons ∷ Vec a (S n) → (a, Vec a n)
uncons (x :- xs) = (x, xs)


-- | Test whether a 'Vec' has zero length.
null ∷ Vec a n → 𝔹
null Nil = T
null _   = F


-- | Returns the length (numbers of elements) of a 'Vec'.
length ∷ Vec a n → ℕ
length Nil       = Z
length (_ :- xs) = S (length xs)

-- | Same as 'length' but produces a singleton type 'Sℕ'.
sLength ∷ Vec a n → Sℕ n
sLength Nil       = SZ
sLength (x :- xs) = SS (sLength xs)



-- * Transformations

-- | Applies a function on every element of a 'Vec'.
map ∷ (a → b) → Vec a n → Vec b n
map _ Nil       = Nil
map f (x :- xs) = f x :- map f xs


-- | Reverse the order of the elements of a 'Vec'
-- > reverse ∘ reverse = id
reverse ∷ Vec a n → Vec a n
reverse Nil       = Nil
reverse (y :- ys) = gcastWith proof (reverse ys ⧺ (y :- Nil))
  where proof = succ_plus (sLength ys)


-- | 'intersperse' @s@ takes a 'Vec' and produces one where
-- @s@ is interspersed (ie inserted between) every two elements
-- of the vector
intersperse ∷ a → Vec a n → Vec a (n :+ Pred n)
intersperse _ Nil        = Nil
intersperse _ (x :- Nil) = x :- Nil
intersperse s (x :- xs)  = gcastWith proof (x :- s :- intersperse s xs)
  where
    n     = sLength xs
    proof = double (SS n)


-- | 'intercalate' @xs xss@ is equivalent to
-- @('concat' ('intersperse' xs xss))@. It inserts the vector @xs@
-- in between the vectors in @xss@ and concatenates the result.
intercalate ∷ Vec a n → Vec (Vec a n) m → Vec a ((m :+ Pred m) :× n)
intercalate xs xss = concat (intersperse xs xss)


-- | The 'transpose' function transposes the rows and columns of its argument.
-- For example,
--
-- > transpose [[1,2,3],[4,5,6]] == [[1,4],[2,5],[3,6]]
--
transpose ∷ SingI n ⇒ Vec (Vec a n) m → Vec (Vec a m) n
transpose Nil         = replicate Nil
transpose (xs :- xss) = gcastWith proof (zipWith (:-) xs (transpose xss))
  where proof = min_self (sLength xs)


-- | 'foldl' applied to a binary operator, a starting value
-- and a 'Vec' reduces the vector to a single value obtained
-- by sequentially applying the operation from the left to the right.
foldl ∷ (a → b → a) → a → Vec b n → a
foldl _ x Nil       = x
foldl (⊗) x (y :- ys) = foldl (⊗) (x ⊗ y) ys


-- | Same as "fold" but reduces the vector in the opposite
-- direction: from the right to the left.
foldr ∷ (a → b → b) → b → Vec a n → b
foldr _ x Nil     = x
foldr (⊗) x (y:-ys) = y ⊗ (foldr (⊗) x ys)


-- | Variant of 'foldl' which requires no starting value but
-- applies only on nonempty 'Vec'
foldl₁ ∷ (a → a → a) → Vec a (S n) → a
foldl₁ _   (x :- Nil)     = x
foldl₁ (⊗) (x :- y :- ys) = foldl₁ (⊗) (x ⊗ y :- ys)


-- | As for 'foldl', a variant of 'foldr' with no starting point
foldr₁ ∷ (a → a → a) → Vec a (S n) → a
foldr₁ _   (x :- Nil)     = x
foldr₁ (⊗) (x :- y :- ys) = x ⊗ (foldr₁ (⊗) (y :- ys))



-- * Special folds

-- | The concatenation of all the elements of a vector of vectors.
concat ∷ Vec (Vec a n) m → Vec a (m :× n)
concat Nil = Nil
concat (xs :- xss) = gcastWith proof (xs ⧺ concat xss)
  where
    n = sLength xs
    m = sLength xss
    proof = plus_commut n (m %:× n)


-- | Map a function over all the elements of a vector and concatenate
-- the resulting vectors.
concatMap ∷ (a → Vec b n) → Vec a m → Vec b (m :× n)
concatMap f = concat ∘ map f


-- | Applied to a a value produces a vector obtained by
-- duplicating the value @n@ times.
replicate ∷ ∀ a n. SingI n ⇒ a → Vec a n
replicate = replicate' (sing ∷ Sℕ n)


-- | 'replicate' variant with an explicit length argument.
replicate' ∷ Sℕ n → a → Vec a n
replicate' SZ     _ = Nil
replicate' (SS n) a = a :- replicate' n a


-- | 'and' returns the conjunction of a container of bools.
and ∷ Vec 𝔹 (S n) → 𝔹
and = foldr₁ (∧)


-- | 'or' returns the disjunction of a container of bools.
or ∷ Vec 𝔹 (S n) → 𝔹
or = foldr₁ (∨)


-- | Determines whether any element of the structure satisfies the predicate.
any ∷ (a → 𝔹) → Vec a n → 𝔹
any _ Nil       = F
any p (x :- xs) = or (map p (x :- xs))


-- | Determines whether all elements of the structure satisfy the predicate.
all ∷ (a → 𝔹) → Vec a n → 𝔹
all _ Nil       = T
all p (x :- xs) = and (map p (x :- xs))


-- | The least element of a vector.
minimum ∷ Ord a ⇒ Vec a (S n) → a
minimum = foldr₁ T.min


-- | The largest element of a vector.
maximum ∷ Ord a ⇒ Vec a (S n) → a
maximum = foldr₁ T.max


-- | The 'sum' function computes the sum of the numbers of a vector.
sum ∷ Num a ⇒ Vec a n → a
sum = foldr (T.+) 0


-- | The 'product' function computes the product of the numbers of a vector.
product ∷ Num a ⇒ Vec a n → a
product = foldr (T.*) 0


-- | 'take' returns the first @n@ element of the vector
take ∷ ∀ a n m. SingI n ⇒ Vec a m → Vec a n
take = take' (sing ∷ Sℕ n)


-- | Variant of 'take' with an explicit argument.
take' ∷ Sℕ n → Vec a m → Vec a n
take' SZ      _        = Nil
take' (SS n) (x :- xs) = x :- take' n xs


-- | 'drop' returns every element of the vector but the first @n@
drop ∷ ∀ a n m. SingI n ⇒ Vec a m → Vec a (m :- n)
drop = drop' (sing ∷ Sℕ n)


-- | Variant of 'drop' with an explicit argument.
drop' ∷ Sℕ n → Vec a m → Vec a (m :- n)
drop' SZ     x         = x
drop' (SS n) (x :- xs) = drop' n xs



-- * Zipping and unzipping vectors

-- | 'zip' takes two vectors and returns a vector of corresponding pairs.
-- If one input vectors is short, excess elements of the longer vector are
-- discarded.
--
-- 'zip' is right-lazy:
--
-- > zip Nil ⊥ = Nil
zip ∷ Vec a n → Vec b m → Vec (a, b) (Min n m)
zip = zipWith (,)


-- | 'zipWith' generalises 'zip' by zipping with the function given
-- as the first argument, instead of a tupling function.
-- For example, @'zipWith' (+)@ is applied to two vectors to produce the
-- vector of corresponding sums.
--
-- 'zipWith' is right-lazy:
--
-- > zipWith f Nil ⊥ = Nil
zipWith ∷ (a → b → c) → Vec a n → Vec b m → Vec c (Min n m)
zipWith _ Nil _ = Nil
zipWith _ _ Nil = Nil
zipWith (⊗) (x :- xs) (y :- ys) = x ⊗ y :- zipWith (⊗) xs ys


-- | 'unzip' transforms a vectors of pairs into a vector of first components
-- and a vector of second components.
unzip ∷ Vec (a, b) n → (Vec a n, Vec b n)
unzip Nil             = (Nil, Nil)
unzip ((x, y) :- xys) = (x :- xs, y :- ys)
  where (xs, ys) = unzip xys
